
#include "dk2nu/tree/dkmeta.h"
#include "dk2nu/tree/dk2nu.h"
#include "MakeReweight.h"

#include <string>
#include <vector>
#include <iostream>
#include <stdlib.h>
#include <iomanip>
#include <fstream>

#include "TH1D.h"
#include "TH2D.h"

//Some constants:
const int Ndet  = 3;
const int Nhel  = 2;

const int Nbins        = 101;
const int Nbins_plus_1 = 102;
const double binw      = 0.2; //GeV
//const double emin      =   0.; // unused
const double emax      = 120.;
const char* hel[Nhel]  = {"numu","numubar"};
const char* det[Ndet]  = {"MINOS_ND","NOvA_ND","NOvA_FD"};
//check in dkmetaTree if these indices are right for the g4numi ntuples being used.  
const int idet[Ndet]   = {2,3,9}; 
const double pi    = 3.14159265358979323846;

using namespace NeutrinoFluxReweight;

class nu_g4numi;
class Numi2Pdg;

int idx_hel(int pdgdcode);
//! returns string with "proper" xrootd enabled path if reading from PNFS (dCache)
std::string change_to_xrootd_path(std::string temp);


/*!
 * Run to get the MINOS to NOvA ND/FD transformation matrices. The input is file with a list of ntuples, 
 * the name of the root output file and the option file. In this script, we are storing only 
 * the central value. I recommend to change the number of universes to be used 
 * to 1 or 2: <NumberOfUniverses>100</NumberOfUniverses> in the scripts/inputs_default.xml
 */
void minos2nova(const char* inputFiles, const char* outputFile, const char* optionsFile){ 
  
  //variable binning to catch MINOS and NOvA peaks
  double binning[Nbins_plus_1];
  for(int i=0;i<Nbins;i++)binning[i] = double(i)*binw;
  binning[Nbins] = emax;

  std::cout<< "Instance of MakeReweight()" <<std::endl;
  MakeReweight* makerew = MakeReweight::getInstance();
  makerew->SetOptions(optionsFile); 
  
  std::cout<<"Making an output file to store histograms"<<std::endl;
  TFile* fOut = new TFile(outputFile,"recreate");
  std::cout<<"File name: "<<fOut->GetName()<<std::endl;
  
  int Nuniv = makerew->GetNumberOfUniversesUsed();
  std::cout<<"Nuniverses "<<Nuniv<<std::endl;  
  
  //Histograms:
  TH1D* hflux[Nhel][Ndet];
  TH2D* hwgts[Nhel][Ndet-1];
  TH2D* htrans[Nhel][Ndet-1];
  for(int i=0;i<Nhel;i++){
    for(int j=0;j<Ndet;j++){
      hflux[i][j] = new TH1D(Form("hflux_pppfxcv_%s_%s",det[j],hel[i]),"",Nbins,binning); //this is to check the tranformation
      if(j!=0){
	hwgts[i][j-1]  = new TH2D(Form("hwgts_ppfxcv_%s_vs_%s_%s"  ,det[0],det[j],hel[i]),"",Nbins,binning,Nbins,binning);
	htrans[i][j-1] = new TH2D(Form("htrans_ppfxcv_%s_to_%s_%s" ,det[0],det[j],hel[i]),"",Nbins,binning,Nbins,binning);
      }
    }
  }
    
  //Loading ntuples:
  TChain* chain_evts   = new TChain("dk2nuTree");  
  TChain* chain_meta   = new TChain("dkmetaTree");  
  bsim::Dk2Nu*  dk2nu  = new bsim::Dk2Nu;  
  bsim::DkMeta* dkmeta = new bsim::DkMeta;
  
  std::cout<<" Adding ntuple from: "<<inputFiles<<std::endl;
  
  std::ifstream* fileIn = new std::ifstream(inputFiles);
  std::string line;
  int count_files = 0;
  if(fileIn->is_open()){
    while(getline(*fileIn,line)){
      if(line.find(".root")<2000){
	line = change_to_xrootd_path(line);
	chain_evts->Add(line.c_str());
	std::cout<<"File In: "<<line<<std::endl;
	count_files++;
	if(count_files==1)chain_meta->Add(line.c_str());
      }
    }
  }
  chain_evts->SetBranchAddress("dk2nu",&dk2nu);
  int nentries  = chain_evts->GetEntries();
  chain_meta->SetBranchAddress("dkmeta",&dkmeta);
  chain_meta->GetEntry(0); //all entries are the same     
 
  std::cout<<"N of entries: "<<nentries<<std::endl;
  
  double flxwgt[Ndet] = {0,0,0};
  double enu[Ndet] = {0,0,0};
  
  for(int ii=0;ii<nentries;ii++){  

    if(ii%100000==0)std::cout<<ii/1000<<" k evts"<<std::endl;
    chain_evts->GetEntry(ii); 
    int nuidx = idx_hel(dk2nu->decay.ntype);   
    if(nuidx<0)continue;
    
    makerew->calculateWeights(dk2nu,dkmeta);    
    double cvwgt = makerew->GetCVWeight();    
    for(int kk=0;kk<Ndet;kk++){
      flxwgt[kk] = ( (dk2nu->nuray)[idet[kk]].wgt )*(dk2nu->decay.nimpwt)/pi;	// 3.1416;
      enu[kk]    =   (dk2nu->nuray)[idet[kk]].E;
      if(flxwgt[kk]<0 || enu[kk]<0)std::cout<<"=> Wrong enu, wgt values"<<std::endl; 
      
      hflux[nuidx][kk]->Fill(enu[kk],flxwgt[kk]*cvwgt);
      if(kk!=0){
	hwgts[nuidx][kk-1]->Fill(enu[0],enu[kk],flxwgt[kk]*cvwgt);
      }
    }//end of detectors
  } //end of entries
  
  //Calculating the MINOS to NOvA matrix:
  for(int i=0;i<Nhel;i++){
    for(int j=1;j<Ndet;j++){
      for(int p=1;p<=Nbins;p++){
	for(int k=1;k<=Nbins;k++){
	  double MINOS_cont = hflux[i][0]->GetBinContent(k);	  
	  double wgts_cont  = hwgts[i][j-1]->GetBinContent(k,p);	  
	  if(MINOS_cont>0)htrans[i][j-1]->SetBinContent(k,p,wgts_cont/MINOS_cont);
	  else std::cout<<"Wrong values: MINOS_cont, wgts_cont: "<<MINOS_cont<<", "<<wgts_cont<<" and type: "
			<<hel[i]<<", det: "<<det[j]<<", binY: "<<p<<", binX: "<<k<<std::endl;
	}
      }//end of bins
    }//end of det
  }//end of hel

// array bounds are ‘TH2D* [2]’ - presume that htrans[i][0] and htrans[i][1] are intended
  std::cout<<"storing general histos"<<std::endl;
  fOut->cd();
  for(int i=0;i<Nhel;i++){
    htrans[i][0]->Write();
    htrans[i][1]->Write();
    fOut->mkdir(Form("h%s",hel[i]));
    fOut->cd(Form("h%s",hel[i]));
    for(int j=0;j<Ndet;j++){
      hflux[i][j]->Write();	
      if(j!=0){
	hwgts[i][j-1]->Write();
      }
    }
  }
  
  //Releasing memory:
  makerew->resetInstance();
  
  std::cout<<"End of run()"<<std::endl;

}

std::string change_to_xrootd_path(std::string temp){
//  pnfs access via root streaming is now not OK. Must switch to xrootd. Dec 2021 -Pierce Weatherly
//  std::string input_flux_dir_fe =  "/pnfs/dune/"+disk+"/users/"+input_user+"/fluxfiles/g4lbne/"+version+"/"+physics_list+"/"+macro+"/"+current+"/flux/";
//  std::string rootxdstr = "root://fndca1.fnal.gov:1094/pnfs/fnal.gov/usr/dune/";
//  std::string input_flux_dir =  rootxdstr+disk+"/users/"+input_user+"/fluxfiles/g4lbne/"+version+"/"+physics_list+"/"+macro+"/"+current+"/flux/";
//  std::cout<<"\n\t Using input directory: "<<input_flux_dir<<std::endl;
//  if(on_grid) {
//    input_flux_dir = getenv("_CONDOR_SCRATCH_DIR");
//    input_flux_dir += "/";
//    std::cout<<"Running on grid, so getting flux files from local disk: "<<input_flux_dir<<std::endl;
//  }

  // pnfs - dCache access via XROOTD logic
  bool enable_xrootd = true;
  std::string temp_inputFile = temp;
  std::string rootxdstr = "root://fndca1.fnal.gov:1094/pnfs/fnal.gov/usr/";
  std::size_t index = temp_inputFile.find("/pnfs/");
  if( index == std::string::npos || index != 0) enable_xrootd = false; // cannot find /pnfs/
  else if(index == 0) enable_xrootd = true; // if the path starts with /pnfs/ , then go ahead and use xrootd.
//  if(enable_xrootd && temp_inputFile.find(grid_dir) != string::npos) enable_xrootd = false;  // if on-grid, leave alone
//  if(enable_xrootd && temp_inputFile.find("/data/") != string::npos) enable_xrootd = false;  // if grabing from dune/data or whatever, leave along; make sure not in pnfs
 // if(enable_xrootd && temp_inputFile.find("root:/") != string::npos) enable_xrootd = false;  // if it looks like it is xrootd already
  if(enable_xrootd){
        temp_inputFile.replace(0,6,rootxdstr);
        printf("\n Replacing string for inputFIle %s\n\t with XROOTD path: %s\n", temp.c_str(), temp_inputFile.c_str() );
  }else printf("\n Input File not on /pnfs/ (dCache); File path: %s\n", temp_inputFile.c_str() );
  return temp_inputFile;
}

  
int idx_hel(int pdgcode){
  int idx = -1;
  if(pdgcode ==  14)idx = 0;
  if(pdgcode == -14)idx = 1;
  return idx;
}

//////////////

#ifndef __CINT__
int main(int argc, const char* argv[]){
  
  minos2nova(argv[1],argv[2],argv[3]);
  return 0;
}
#endif
